using System;
using System.Data;
using System.Data.SqlClient;

using Westwind.Tools;

namespace FoxWebServiceInterop
{
	/// <summary>
	/// Summary description for SimpleBusObj.
	/// </summary>
	public class SimpleBusObj
	{
		
		protected SqlConnection Conn;
		
		/// <summary>
		/// Public external reference to the DataSet for this bus object
		/// </summary>
    	public DataSet DataSet;

		/// <summary>
		/// Public external reference of the DataRow that holds the result
		/// from Load() or like operations. Table name for this operation
		/// should always be this.TableName + "_Record" pointing at row 0
		/// </summary>
		public DataRow DataRow;


		public string ConnectionString;
		public string TableName = "";
		public string PkField = "Pk";
		public string ErrorMsg = "";
		public bool Error = false;

		public bool ExecuteWithSchema = false;
		
		public  SimpleBusObj()
		{
			//
			// TODO: Add constructor logic here
			//
		}

		/// *********************
		/// DATA UTILITY ROUTINES	
		/// *********************
		/// these are not exposed throught the Web Service
		/// and really belong into an external class.	


		/// <summary>
		/// Sets up the connection for a query
		/// </summary>
		/// <returns></returns>
		protected virtual bool Open()
		{
			/// create if it doesn't exist already
			if (this.Conn == null) 
			{
				try 
				{
					this.Conn = new SqlConnection(this.ConnectionString);
				}
				catch(Exception e)
				{
					this.SetError( e.Message );
					return false;
				}
			}

			/// check if connection is open - if not open it
			if (this.Conn.State != ConnectionState.Open) 
			{
				try 
				{
					Conn.Open();
				}
				catch(Exception e)
				{
					this.SetError( e.Message );
					return false;
				}
			}

			/// make sure our dataset object exists
			if (this.DataSet == null) 
				this.DataSet = new DataSet();

			return true;
		}

		protected virtual bool Close() 
		{
			if (Conn.State == ConnectionState.Open)
				try 
				{
					Conn.Close();
				}
				catch(Exception e)
				{
					this.SetError( e.Message );
					return false;
				}

			return true;
		}

		/// <summary>
		///  Executes a Select statement and sets teh value of a the dataRowSet.
		/// </summary>
		/// <param name="lcSQL"></param>
		/// <param name="lcCursor"></param>
		/// <returns></returns>
		public virtual int Execute(string SQL, string Cursor)
		{
			return Execute(SQL, Cursor,null);
		}

		/// <summary>
		///  Executes a Select statement and sets teh value of a the dataRowSet.
		/// </summary>
		/// <param name="SQL">Sql string to execute</param>
		/// <param name="Cursor">Name of the resulting table in the DataSet property</param>
		/// <param name="Parameters">variable number of Sql Parameter object for this query</param>
		/// <returns></returns>
		public virtual int Execute(string SQL, string Cursor, params SqlParameter[] Parameters)
		{
			if (!this.Open()) 
				return 0;

			SqlDataAdapter Adapter = new SqlDataAdapter();
			if (this.ExecuteWithSchema)
				Adapter.MissingSchemaAction = MissingSchemaAction.AddWithKey;

			Adapter.SelectCommand = new SqlCommand(SQL,Conn);

			if (Parameters != null) 
			{
				foreach(SqlParameter Parm in Parameters) 
					Adapter.SelectCommand.Parameters.Add( Parm );
			}

			if ( DataSet.Tables.Contains(Cursor) )
					DataSet.Tables.Remove(Cursor);
	
			try 
			{
				Adapter.Fill(this.DataSet,Cursor);
			}
			catch(Exception e)
			{
				this.SetError(e.Message);
				this.Close();
				return -1;
			}
			
			this.Close();

			return DataSet.Tables[Cursor].Rows.Count;
		}

		// *** HIGHER LEVEL METHODS (very, very basic)

		public virtual bool Load(string Pk)
		{
			if ( this.Execute("select * from " + this.TableName + " where " + this.PkField + "='" + Pk + "'",this.TableName + "_Record") < 1)
				return false;

			this.DataRow = this.DataSet.Tables[this.TableName + "_Record"].Rows[0];

			return true;
		}


		public virtual bool Load(int Pk) 
		{
			if ( this.Execute("select * from " + this.TableName + " where " + this.PkField + "=" + Pk.ToString(),this.TableName + "_Record") < 0)
				return false;

			this.DataRow = this.DataSet.Tables[this.TableName + "_Record"].Rows[0];

            return true;
		}

		public virtual bool New() 
		{
			if (!this.GetBlankRecord())
				return false;

			return true;
		}

		public virtual bool GetBlankRecord() 
		{
			if (this.Execute("select * from " + this.TableName + " where 0=1",this.TableName + "_Record" ) < 0)
				return false;

			// *** Add a blank row - all null values
			this.DataSet.Tables[this.TableName + "_Record"].Rows.Clear();
			DataRow NewRow = this.DataSet.Tables[this.TableName + "_Record"].NewRow();

			this.DataSet.Tables[this.TableName + "_Record"].Rows.Add(NewRow);

			this.DataRow = this.DataSet.Tables[this.TableName + "_Record"].Rows[0];
			wwDataUtils.InitializeDataRowWithBlanks(this.DataRow);

			return true;
		}

		public bool Save() 
		{
			return Save(this.TableName + "_Record");
		}
		public bool Save(string Table) 
		{
			if (!this.Open())	
				return true;

			/// Even though we won't execute this command the data adapter needs this 
			/// command to figure out the key field info in order to properly do updates
			SqlDataAdapter oAdapter = new SqlDataAdapter("select * from " + this.TableName,this.Conn);

			/// This builds the Update/InsertCommands for the data adapter
			SqlCommandBuilder oCmdBuilder = new SqlCommandBuilder(oAdapter);

			int lnRows = 0;
			try 
			{
				/// Take the changes in the dataset and sync to the database
				lnRows = oAdapter.Update(this.DataSet,Table);
			}
			catch(Exception ex)
			{
				this.SetError(ex.Message);
				return false;
			}
			finally 
			{
				this.Close();
			}

			if ( lnRows < 1 )
				return false;

			return true;
		}

		
		
		protected void SetError(string lcErrormsg)
		{
			if (lcErrormsg.Length == 0) 
			{
				this.ErrorMsg = "";
				this.Error = false;
				return;
			}

			this.ErrorMsg = lcErrormsg;
			this.Error = true;
		}

	}
}
