using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Web;
using System.Web.Services;

using System.IO;
using System.Web.Services.Protocols;
using System.Xml;
using System.Reflection;

using Westwind.Tools;
//using DataRowContainers;

namespace FoxWebServiceInterop
{
	/// <summary>
	/// Summary description for FoxInteropService.
	/// </summary>
	[WebService(Namespace="http://www.west-wind.com/FoxWebServiceInterop/",	        
		        Description="Small Sample Service that demonstrates interaction between .Net Web Services and Visual FoxPro")]
	public class FoxInteropService : System.Web.Services.WebService
	{
		public FoxInteropService()
		{
			InitializeComponent();
		}

		#region Component Designer generated code
		
		//Required by the Web Services Designer 
		private IContainer components = null;
				
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if(disposing && components != null)
			{
				components.Dispose();
			}
			base.Dispose(disposing);		
		}
		
		#endregion

		// WEB SERVICE EXAMPLE
		// The HelloWorld() example service returns the string Hello World
		// To build, uncomment the following lines then save and build the project
		// To test this web service, press F5

		[WebMethod(Description="The Quintessential Hello World Method")]
		public string HelloWorld(string Name)
		{
			return "Hello World " + Name + 
                   "! The time is: " + DateTime.Now.ToString();
		}

		[WebMethod(Description="Returns the Server Time")]
		public DateTime GetServerTime() 
		{
			return DateTime.Now;
		}

#if true
		[WebMethod(Description="Runs a generic Query on the server and returns the result as an XML string")]
		public string RunQueryXml(string Sql,string TableName)
		{
			busAuthor Author = new busAuthor();

			Author.ExecuteWithSchema = true;

			// *** Execute populates a table called Authors in Author.DataSet
			if ( Author.Execute(Sql,TableName) < 0)
				throw new SoapException(Author.ErrorMsg,SoapException.ServerFaultCode);

			StringWriter Writer = new StringWriter();
			Author.DataSet.WriteXml(Writer,XmlWriteMode.WriteSchema);
			return Writer.ToString();
		}


		[WebMethod(Description="Runs a generic Query on the server and returns the result as a DataSet")]
		public DataSet RunQuery(string Sql,string TableName)
		{
			busAuthor Author = new busAuthor();

			Author.ExecuteWithSchema = true;

			// *** Execute populates a table called Authors in Author.DataSet
			if ( Author.Execute(Sql,TableName) < 0)
				throw new SoapException(Author.ErrorMsg,SoapException.ServerFaultCode);

			return Author.DataSet;
		}
#endif

#if true
		/// <summary>
		/// Returns an author dataset as a query
		/// </summary>
		/// <param name="Name"></param>
		/// <returns></returns>
		[WebMethod(Description="Returns an Author table in a DataSet object")]
		public DataSet GetAuthors(string Name) 
		{
			busAuthor Author = new busAuthor();
			
			// *** Force Complete Schema to be loaded so we get proper type info
			Author.ExecuteWithSchema = true;

			int Result = Author.GetAuthors(Name);

			if (Author.Error) 
				throw new SoapException(Author.ErrorMsg,SoapException.ServerFaultCode);

			if (Result < 0) 
			{
				return null;
			}

			return Author.DataSet;
		}


		[WebMethod(Description = "Updates the Authors Table from a table called Authors passed in as part of a DataSet")]
		public bool UpdateAuthorData(DataSet Ds)
		{
			busAuthor Author = new busAuthor();
			
			Author.DataSet = Ds;

			bool Result = Author.Save("Authors");
			if (!Result)
				throw new SoapException(Author.ErrorMsg,SoapException.ServerFaultCode);

			return Result;
		}

		/// <summary>
		/// Updates the Dataset with an XML Diffgram
		/// </summary>
		/// <param name="lcXml">Diffgram of all changes made from VFP Code</param>
		/// <returns></returns>
		[WebMethod(Description = "Updates the Authors Table from a table called Authors passed in as part of a DataSet using an XML string.")]
		public bool UpdateAuthorDataXml(string lcXml) 
		{
			DataSet Ds = new DataSet();
			Ds.ReadXml(new StringReader(lcXml),XmlReadMode.DiffGram);
			bool Result = UpdateAuthorData(Ds);
			return Result;
		}


		[WebMethod(Description = "Retrieves a list of author entity objects as an array result.")]
		public AuthorEntity[] GetAuthorArray(string Name) 
		{
			DataSet ds = this.GetAuthors(Name);
			if (ds == null)
				return null;

			AuthorEntity Authors = new AuthorEntity();
			ArrayList AuthorList = Authors.CreateEntityList(ds.Tables[0]);

			return (AuthorEntity[]) AuthorList.ToArray( typeof( AuthorEntity)  );
		}

		/// <summary>
		/// Returns a list of authors for list display 
		/// </summary>
		/// <returns></returns>
		[WebMethod(Description = "Returns a list of authors for display in a ListBox with id,lname and fname in a DataSet and a table called AuthorList.")]
		public DataSet GetAuthorsList() 
		{
			busAuthor Author = new busAuthor();
			Author.ExecuteWithSchema = true;

			int Result = Author.GetAuthors("","au_id,au_lname,au_fname","AuthorList");

			if (Author.Error) 
				throw new SoapException(Author.ErrorMsg,SoapException.ServerFaultCode);

			if (Result < 0) 
				return null;

			return Author.DataSet;
		}

		/// <summary>
		/// Returns a single author from the Web Service
		/// </summary>
		/// <param name="ID"></param>
		/// <returns></returns>
		[WebMethod(Description = "Returns an individual author by Id as an Authors Table in DataSet.")]
		public DataSet GetAuthor(string ID) 
		{
			busAuthor Author = new busAuthor();
			Author.ExecuteWithSchema = true;

			if ( !Author.Load(ID) )
				throw new SoapException(Author.ErrorMsg,SoapException.ServerFaultCode);

			// *** Must return entire Data Set
			// *** will be in the 'Authors' table
			return Author.DataSet;
		}


		/// <summary>
		/// Retrieves an Author business object and returns an Entity object
		/// which is more light weight than a dataset for a single record
		/// </summary>
		/// <param name="Id"></param>
		/// <returns></returns>
		[WebMethod(Description = "Returns an individual Author by ID as an Entity Object.")]
		public AuthorEntity GetAuthorEntity(string Id) 
		{
			busAuthor Author = new busAuthor();

			if (!Author.Load(Id))
				return null;

			AuthorEntity Auth = new AuthorEntity();
			Auth.LoadRow(Author.DataRow);

			return Auth;
		}


		[WebMethod(Description = "Updates or adds an individual Author from a passed in AuthorEntity object.")]
		public bool UpdateAuthorEntity(AuthorEntity UpdatedAuthor) 
		{
			if (UpdatedAuthor == null)
				return false;

			busAuthor Author = new busAuthor();
	
			if (!Author.Load(UpdatedAuthor.Au_id))  
			{
				if (!Author.New()) 
					return false;
			}

			UpdatedAuthor.SaveRow(Author.DataRow);

			if (!Author.Save())
				return false;

			return true;
		}




		[WebMethod(Description = "Returns Authors DataSet as an XML string.")]
		public string GetAuthorsXml(string Name) 
		{
			DataSet Ds = this.GetAuthors(Name);
			if (Ds == null)
				return "";
		
			StringWriter Writer = new StringWriter();
			Ds.WriteXml(Writer,XmlWriteMode.WriteSchema);
			return Writer.ToString();
		}


		[WebMethod(Description="Intentional Exception test")]
		public void ExceptionTest() 
		{
			throw new SoapException("Intentional Exception Thrown",SoapException.ServerFaultCode);
		}

		[WebMethod(Description = "Demonstrates returning an XML node")]
		public XmlNode ReturnXml()
		{
			XmlDocument Dom = new XmlDocument();
			Dom.LoadXml("<x><y>Data</y><y>More Data</y></x>");

			return Dom.DocumentElement;
		}

		[WebMethod]
		public byte[] ReturnBytes() 
		{
			return new byte[5] { 48, 49, 0 , 48, 49 };
		}
#endif
	}
}
